#include "Arduboy.h"
#include "gfx.h"
#include "music.h"
#include "level.h"
#include "ab_logo.c"

#define TVOUT 0
#define levHeight 9
#define levWidth 14

#define X_MAX = 128-14
#define Y_MAX = 64-15-8
#define GRAVITY 0.25
#define joeJumpHeight 2.5
#define frameRate 30
#define stepSize 1
#define DOORTILE 6
#define ITEMTILE 1

#define SKIPINTRO 1
#define OUTPUTVIDEO 1

Arduboy arduboy;

unsigned char * screenBuffer;
byte sfxPos = 0;
bool playSfx = false;
const byte *SFX = 0;
long int frameNumber = 0;
byte contrast = 0;
int FPS = 0;

byte joe_frame = 0;
signed char joeX, joeY = 0;
float joeFallSpeed;
byte input; // for capturing buttons
bool dimension = 0;
bool oldDimension = 1;
byte joeStep = 0;
byte roomNumber = 0;
bool joeDirection = 0; // left
byte trkNum = 0;
bool flicker = 0;
byte gameMode = 0;
byte oldGameMode = 0;
byte joeFrame = 0;
int waterFrame=0;

// items/keys etc.
const byte PROGMEM keys[8]= { 0,12,20,10, 6, 0, 0, 0}; // rooms that the keys are in
const byte PROGMEM locks[8]={ 9, 0, 0,11,12, 0, 0, 0}; // rooms for those keys
byte gotKeys[8];

// enemies!
byte numEnemies;
byte enemyX[5]; // We'll max out a 5 enemies per screen
byte enemyY[5];
signed char enemyDir[5];
byte enemyFrame[5];

// level re-entry!
int levSX=1; // start x
int levSY=1; // start y
byte levSD=0; // start dimension
byte levSDir=0; // start direction
float levFS=0; // fall speed

byte menuItem = 0;
#define _SFX 0
#define _MUSIC 1
byte settings[5]; // settings

// Button updates
#define HELD 0
#define NEW 1
#define RELEASE 2
byte CompletePad, ExPad, TempPad;
bool _A[3], _B[3], _Up[3], _Down[3], _Left[3], _Right[3];

void UPDATEPAD(int pad, int var) {
  _B[pad] = (var & B_BUTTON) >> 2; // 4
  _A[pad] = (var & A_BUTTON) >> 3; // 8
  _Down[pad] = (var & DOWN_BUTTON) >> 4; // 16
  _Left[pad] = (var & LEFT_BUTTON) >> 5; // 32
  _Right[pad] = (var & RIGHT_BUTTON) >> 6; // 64
  _Up[pad] = (var & UP_BUTTON) >> 7; // 128
}
void UpdatePad(int joy_code)
{
  ExPad = CompletePad;
  CompletePad = joy_code;
  UPDATEPAD(HELD, CompletePad); // held
  UPDATEPAD(RELEASE, (ExPad & (~CompletePad))); // released
  UPDATEPAD(NEW, (CompletePad & (~ExPad))); // newpress
}


// print text
void print(char x, char y, const char* text, bool color = 1){
  for(char t = 0; t < strlen(text); t++){
    char character = text[t]-32;
    arduboy.drawBitmap(x,y,myFont+(character*3),3,6,color);
    x += 4;
  }
}

static inline void fps(const int seconds){
  // Create static variables so that the code and variables can
  // all be declared inside a function
  static unsigned long lastMillis;
  static unsigned long frameCount;
  static unsigned int framesPerSecond;
  
  // It is best if we declare millis() only once
  unsigned long now = millis();
  frameCount ++;
  if (now - lastMillis >= seconds * 1000) {
    framesPerSecond = frameCount / seconds;
    //Serial.println(framesPerSecond);
    FPS = framesPerSecond;
    frameCount = 0;
    lastMillis = now;
  }
}


// change brightness?
void setContrast(uint8_t contrast)
{
  arduboy.LCDCommandMode();
  SPI.transfer(0x81); // contrast command
  SPI.transfer(contrast);
  arduboy.LCDDataMode();
}

void fadeOut() {
  // set screen to funky settings to allow fade
  arduboy.LCDCommandMode();
  SPI.transfer(0xD9); // charge time command
  SPI.transfer(0x00); // 0 - 16
  SPI.transfer(0x81); // contrast command
  SPI.transfer(127);
  SPI.transfer(0xdb); // VCOM deselect level mode
  SPI.transfer(0x00); // set vcomh = 0.83 * VCC
  SPI.transfer(0xa4); // set entire display on/off
  SPI.transfer(0xa6); // set normal display
  SPI.transfer(0xaf); // set display on
  arduboy.LCDDataMode();  

  for (int t = 127; t > 0; t -= 1) {
    setContrast(t);
    arduboy.display();
    playSFX(); // make sure sfx play every chance!
  }
  setContrast(0);
}
void fadeIn() {
  for (int t = 0; t < 127; t += 1) {
    setContrast(t);
    arduboy.display();
    UpdatePad(arduboy.buttonsState());
    playSFX(); // make sure sfx play every chance!
  }
  // restore screen settings to default
  arduboy.LCDCommandMode();
  SPI.transfer(0xD9); // charge time command
  SPI.transfer(0xF1); // 0 - 16
  SPI.transfer(0x81); // contrast command
  SPI.transfer(0xCF);
  SPI.transfer(0xdb); // VCOM deselect level mode
  SPI.transfer(0x40); // set vcomh = 0.83 * VCC
  SPI.transfer(0xa4); // set entire display on/off
  SPI.transfer(0xa6); // set normal display
  SPI.transfer(0xaf); // set display on
  arduboy.LCDDataMode();  
}

void drawBorders(){
  // draw black borders down the left+right of the screen
  for(char y = 0; y<8; y++){
    if(dimension){
      drawTile(0, y, ground+15*8);
      drawTile(15, y, ground+15*8);
    }else{
      drawTile(0, y, ground);
      drawTile(15, y, ground);
    }
  }
  
}

void titleScreen() {
  dimension = 1;
  roomNumber = 0;
  joeX = 8;
  joeY = 0;
  for(byte t=0; t<8; t++){
    gotKeys[t]=0;
  }

  if (_Up[NEW] && menuItem >= 1) {
    menuItem--;
  }
  if (_Down[NEW]) {
    menuItem++;
  }

  if (_B[NEW]) {
    if (menuItem == 0) {
      arduboy.setFrameRate(frameRate);
      playSfx = false;
      sfxPos = 0;
      trkNum = 0; // reset music
      joeDirection = 0; // left
      gameMode = 10;
      readLevel();
    }
    if (menuItem == 1) {
      settings[_SFX] = 1 - settings[_SFX];
    }
    if (menuItem == 2) {
      settings[_MUSIC] = 1 - settings[_MUSIC];
    }
  }

  if (menuItem > 2) {
    menuItem = 2;
  }

  arduboy.clear();

  for(char y=0; y<8; y++){
    for(char x=0; x<15; x++){
      int mt = pgm_read_byte(titleMap+(x+16*y));
      drawTile(x,y, titleTiles + (mt*8));
    }
  }
  // tagline
  print(7,34,"     AND");
  print(7,42,"THE MAGIC HAT");
  
  drawTile(8,2+2*menuItem, cursorSprite);
  
  if (settings[_MUSIC]) {
    print(74,50,"O");
  }
  if (settings[_SFX]) {
    print(74,34,"O");
  }

  char text[] = "        ";
  sprintf(text, "FPS:%d",FPS);
  print(0, 0 ,text);
  arduboy.display();
}

void drawTile(int16_t x, int16_t y, const uint8_t *bitmap) {
  for (byte t = 0; t < 8; t++) {
    screenBuffer[(x * 8) + (y * 128) + t] = pgm_read_byte(bitmap + t);
  }
}

void drawTileOffset(int16_t x, int16_t y, const uint8_t *bitmap) {
  x+=1;
  for (byte t = 0; t < 8; t++) {
    if(y>0){
      byte temp = screenBuffer[(x * 8) + ((y-1) * 128) + t];
      byte myByte = (pgm_read_byte(bitmap + t)&0x0F)<<4;
      screenBuffer[(x * 8) + ((y-1) * 128) + t] += myByte;
    }
    if(y<8){
      byte temp = screenBuffer[(x * 8) + (y * 128) + t];
      byte myByte = (pgm_read_byte(bitmap + t)&0xF0)>>4;
      screenBuffer[(x * 8) + (y * 128) + t] += myByte;
    }
  }
}

void readLevel(){
  // read through current level, find enemy start points etc.
  numEnemies = 0; // reset enemy numbers
  for (char y = 0; y < levHeight; y++) {
    for (char x = 0; x < levWidth; x++) {
      char tile = pgm_read_byte(&level[(x + levWidth * y) + (roomNumber * ((levWidth*levHeight) + 4))]);
      if (dimension) {
        tile &= 0x0F;
      } else {
        tile = (tile & 0xF0) >> 4;
      }
      switch(tile){
        case 4:
          enemyX[numEnemies] = x*8;
          enemyY[numEnemies] = y*8;
          enemyDir[numEnemies] = 1;
          enemyFrame[numEnemies]= 0;
          numEnemies++;
        break;
      }
    }
  }
}

void intro(){
  arduboy.invert(false);
  arduboy.tunes.stopScore();
  arduboy.clear();
  gameMode = 1;
}



void setup() {
  //arduboy.begin();
  arduboy.beginNoLogo();
  arduboy.setFrameRate(frameRate);
  // arduboy.clear();
  screenBuffer = arduboy.getBuffer();
  arduboy.initRandomSeed();
  gameMode = 0;

  joeFallSpeed = 0;
  joeX = 8;
  joeStep = 0;
/*
  arduboy.LCDCommandMode();

  SPI.transfer(0xD9); // charge time command
  SPI.transfer(0x00); // 0 - 16
  SPI.transfer(0x81); // contrast command
  SPI.transfer(127);

  SPI.transfer(0xdb); // VCOM deselect level mode
  SPI.transfer(0x00); // set vcomh = 0.83 * VCC
  SPI.transfer(0xa4); // set entire display on/off
  SPI.transfer(0xa6); // set normal display
  SPI.transfer(0xaf); // set display on

  arduboy.LCDDataMode();
*/
}

void playGame() {
  // keep the music going...
  if (!arduboy.tunes.playing() && settings[_MUSIC] == 1) {
    byte trackToPlay = pgm_read_byte(&tracker[trkNum]);
    arduboy.tunes.playScore(track[trackToPlay]);
    trkNum++;
    if (trkNum >= sizeof(tracker))trkNum = 2;
  }


  updateMap();
  updateEnemies();
  updateJoe();

  if(dimension){
    // instructions for those who need them
    switch (roomNumber) {
      case 0: //    <- 32 chars accross screen ---->
        print(24,32," You can use the D-pad",0);
        print(24,39," to walk.",0);
        break;
      case 1:
        print(0,0,"Sometimes you may need to press",0);
        print(0,7,"the B button",0);
        break;
      case 2:
        print(0,0,"Some places require great skill.",0);
        break;
      case 3:
        print(0,0,"while others need great courage",0);
        break;
      case 4:
        print(0,32,"But that is not all! Sometimes,",0);
        print(0,39,"Joe can use his top-hat powers",0);
        print(0,46,"to change the world! (D-pad UP)",0);
        break;
      case 23:
        print(0,0,"I think you should go the RIGHT",0);
        print(0,7,"way.       (Get it?)",0);
        break;
    }
  }

}

void playSFX() {
  if (playSfx == true && settings[_SFX] == 1) {
    if (pgm_read_byte(&SFX[sfxPos]) != 0xFF) {
      arduboy.tunes.tone(pgm_read_byte(&SFX[sfxPos++]) * 2, 70); // SFX
    } else {
      playSfx = false;
      sfxPos = 0;
    }
  }
}

void looseLife(){
  // Joe doesn't have lives! 
  fadeOut();
  fadeIn();
  // reset Joes position
  joeX = levSX; // start x
  joeY = levSY; // start y
  dimension = levSD; // start dimension
  joeDirection = levSDir; // start direction
  joeFallSpeed = levFS;
  gameMode=10;
}

void loop() {
  // pause render until it's time for the next frame
  if (!(arduboy.nextFrame()))  return;
  frameNumber++;

  //  setContrast((frameNumber/60)&127);

  // capture buttons once per loop only
  UpdatePad(arduboy.buttonsState());
  playSFX(); // make sure sfx play every chance!

  switch (gameMode) {
    case 0:
      intro();
      break;
    case 1:
      titleScreen();
      break;
    case 10:
      // play game
      playGame();
      break;
    case 15:
      // lose life!
      looseLife();
      break;      
  }


 /* 
    char text[] = "               ";
    sprintf(text, "FRM %d",frameNumber);
    arduboy.setCursor(0, 56);
    arduboy.print(text);
  */
  arduboy.display();

  if(TVOUT){
    if(dimension==0){
      Serial.write(screenBuffer, 128 * 8);
    }else{
      // invert the screen data, because we cant send commands.
      for(int t=0; t<128*8; t++){
        char s = ~screenBuffer[t];
        Serial.write(&s,1);
      }
    }
  }
  
  fps(1);
}

char getTile(int x, int y) {
  x /= 8; y /= 8;

  if (y > 8 || y < 0) return 0;
  if (x > 13 || x < 0) return 0;
  
  int roomOffset = (levWidth*levHeight)+4; // should be 130
  char tile = pgm_read_byte(&level[(x + levWidth * y) + (roomNumber * roomOffset)]);

  if (dimension) {
    tile &= 0x0F;
  } else {
    tile = (tile & 0xF0) >> 4;
  }

  // special cases
  if(tile==9){ // lock
    bool gotKey=0;
    for(byte t=0; t<8; t++){
      if(pgm_read_byte(&locks[t])==roomNumber){
        if(gotKeys[t]==1){
          gotKey=1;
        }
      }
    }
    if(gotKey==0){tile=15;} // make lock umpassable
  }

  
  return tile;
}

void updateJoe() {

  bool jumping = 0;
  bool falling = 0;
  joeY += joeFallSpeed;
  joeFallSpeed += GRAVITY;
  if (joeFallSpeed > 5) {
    joeFallSpeed = 5;
  }
  int roomOffset = (levWidth*levHeight)+4; // should be 130

  signed char joeX1 = joeX + 1;
  signed char joeX2 = joeX + 6;
  signed char joeY1 = joeY;
  signed char joeY2 = joeY + 8;

  if (getTile(joeX + 4, joeY + 8) > 10) { // falling
    joeFallSpeed = 0;
    joeY = (joeY >> 3) << 3;
  }
  if ((getTile(joeX + 4, joeY) > 10 ) && joeFallSpeed < 0) { // jumping
    joeFallSpeed = 0;
    joeY = ((joeY+4) >> 3) << 3;
  }

  // lose a life!
  if(joeFallSpeed==0 && ((getTile(joeX + 4, joeY + 4) == 2 || getTile(joeX + 4, joeY + 4) == 3 ))) { // death tile!
    gameMode=15;
  }

  
  // Colllect things!
  
  if(getTile(joeX + 4, joeY +7)==1){
    for(byte t=0; t<8; t++){
      if(pgm_read_byte(&keys[t])==roomNumber){
        gotKeys[t]=1;
        SFX = SFX_Key;
        playSfx = true;
        
      }
    }
  }
  
  // controls
  if ((_Up[NEW]) && getTile(joeX + 4, joeY +7) == DOORTILE) {
    //fadeOut();
    // check for special doors here
    switch (roomNumber) {
      case 16:
        if(joeX/8==1 && joeY/8==5){
          roomNumber=17;
          joeX = 8;
          joeY = 8;
        }
        break;
      case 21:
        if(dimension){
          roomNumber=14;
          joeX = 0;
          joeY = 8*2;
        }
        break;
      case 22:
        if(joeX/8==7 && joeY/8==1){
          roomNumber=6;
          joeX = 7*8;
          joeY = 5*8;
        }
        break;
      case 23:
        if(dimension){
          roomNumber=14;
          joeX = 0;
          joeY = 8*2;
        }
        break;
    }
    
    dimension = 1 - dimension;
    updateMap();
    readLevel();
    //fadeIn();
    SFX = SFX_Warp;
    playSfx = true;
  }

  if (_A[RELEASE]) {
    gameMode = 0; // reset to title, replace with menu!
  }

  if (_B[NEW] && (getTile(joeX+4, joeY+8) > 10)) {
    joeFallSpeed = -joeJumpHeight;
    SFX = SFX_Jump;
    playSfx = true;
  }

  if ((_Right[HELD]) && ((getTile(joeX + 7 + stepSize, joeY + 4) <= 10 ) || joeX >= 120)) {
    if (joeY >= 0 && joeY <= 56) {
      joeX += stepSize; joeStep++; joeDirection = 0;
    }
  }
  if ((_Left[HELD]) && (getTile(joeX - stepSize, joeY+4) <= 10 )) {
    if (joeY >= 0 && joeY <= 56) { // make sure we're on screen when trying to move
      joeX -= stepSize; joeStep++; joeDirection = 1;
    }
  }

  if (joeStep >= 2) {
    joe_frame++;
    joeStep = 0;
  }
  if (joe_frame >= 6) {
    joe_frame = 0;
  }

  if (joeY < 0) {
    roomNumber = pgm_read_byte(&level[(roomNumber * roomOffset) + roomOffset-4]);  // bottom room number
    joeY = 64;
    joeFallSpeed = -joeJumpHeight;
    levSX = joeX; // start x
    levSY = joeY; // start y
    levSD = dimension; // start dimension
    levSDir = joeDirection; // start direction
    levFS = joeFallSpeed;
    readLevel();
  }
  if (joeY > 64) {
    roomNumber = pgm_read_byte(&level[(roomNumber * roomOffset) + roomOffset-3]);  // bottom room number
    joeY = 0;
    levSX = joeX; // start x
    levSY = joeY; // start y
    levSD = dimension; // start dimension
    levSDir = joeDirection; // start direction
    levFS = joeFallSpeed;
    readLevel();
  }
  if (joeX < -3) {
    roomNumber = pgm_read_byte(&level[(roomNumber * roomOffset) + roomOffset-2]);  // left room number
    joeX = 107;
    levSX = joeX; // start x
    levSY = joeY; // start y
    levSD = dimension; // start dimension
    levSDir = joeDirection; // start direction
    levFS = joeFallSpeed;
    readLevel();
  }
  if (joeX > 107) {
    roomNumber = pgm_read_byte(&level[(roomNumber * roomOffset) + roomOffset-1]);  // right room number
    joeX = -3;
    levSX = joeX; // start x
    levSY = joeY; // start y
    levSD = dimension; // start dimension
    levSDir = joeDirection; // start direction
    levFS = joeFallSpeed;
    readLevel();
  }

  if (joeFallSpeed > 0) {
    falling = 1;
  }
  if (joeFallSpeed < 0) {
    jumping = 1;
  }

  if (_Left[RELEASE] || _Right[RELEASE]) {
    joe_frame = 3;
  }

  // falling == 0, jumping == 0
  int theFrame = (joe_frame * 8);
  if (jumping == 1) {
    theFrame = 6 * 8;
  }
  if (falling == 1) {
    theFrame = 7 * 8;
  }

  if (joeDirection == 1) {
    theFrame += 64;
  }

  // draw Joe
  //arduboy.drawBitmap(joeX+joeOffset, joeY, joeSprite+theFrame+32, 16, 16, 0);
  //drawXORBitmap(joeX+8, joeY-4, joeSprite + theFrame, 8, 8, 1);
  arduboy.drawBitmap(joeX+8, joeY-4, joeSprite + theFrame, 8, 8, 2);

  // hat with mask
  bool ho=0;
  if(theFrame==0 || theFrame==64) ho=1;
  //arduboy.drawBitmap(joeX+6+(joeDirection*4), joeY-7+ho, joeHat+8+(joeDirection*16), 8, 8, 0);
  arduboy.drawBitmap(joeX+6+(joeDirection*4), joeY-7+ho, joeHat+(joeDirection*16), 8, 8, 2);

  drawBorders();
}

void updateEnemies(){

  for(int t=0; t<numEnemies; t++){
    if(getTile((enemyX[t]+4)+(enemyDir[t]*4), enemyY[t]+4) > 10 // if side tile solid
      || getTile((enemyX[t]+4)+(enemyDir[t]*4), enemyY[t]+12) <= 10 // if below side tile missing
      ){
      enemyDir[t] = -enemyDir[t]; // reverse direction
    }

    enemyX[t] += enemyDir[t];

    char spriteFrame = (enemyX[t]%10)/5;
    if(enemyDir[t]>0){spriteFrame+=2;}
    arduboy.drawBitmap(enemyX[t]+8, enemyY[t]-4, enemySprite+(spriteFrame*8), 8, 8, 1);
  }
}

void updateMap() {
  //flicker =  1 - flicker;
  arduboy.clear();
  drawBorders();
  for (char y = 0; y < levHeight; y++) {
    for (char x = 0; x < levWidth; x++) {
      char tile = pgm_read_byte(&level[(x + levWidth * y) + (roomNumber * ((levWidth*levHeight) + 4))]);
      if (dimension) {
        tile &= 0x0F;
        arduboy.invert(true);
      } else {
        tile = (tile & 0xF0) >> 4;
        arduboy.invert(false);
      }
      if(tile==DOORTILE){
        drawTileOffset(x, y, ground);
        for(char t=0; t<8; t++){
          char x1 = random(8);
          char y1 = random(8);
          arduboy.drawPixel((x*8)+x1+8,(y*8)+y1-4,1);
        }
      }else if(tile==ITEMTILE){
        bool gotKey=0;
        for(byte t=0; t<8; t++){
          if(pgm_read_byte(&keys[t])==roomNumber){
            if(gotKeys[t]==1){
              gotKey=1;
            }
          }
        }
        if(gotKey==0){
          arduboy.drawBitmap(((x+1)*8)+(cos(frameNumber)), ((y*8)-4)+(sin(frameNumber)), objects, 8, 8, 2);
        }
      }else if(tile==9){ // lock
        bool gotKey=0;
        for(byte t=0; t<8; t++){
          if(pgm_read_byte(&locks[t])==roomNumber){
            if(gotKeys[t]==1){
              gotKey=1;
            }
          }
        }
        if(gotKey==0){
          drawTileOffset(x, y, ground + (tile * 8));
        }
      }else if(tile==3){
        drawTileOffset(x, y, waterTiles + (((x%8)+waterFrame*8)*8));
      }else{
        drawTileOffset(x, y, ground + (tile * 8));
      }
      //arduboy.drawBitmap((x*8)+8,(y*8)-4,ground + (tile * 8),8,8,1);
    }
  }


  if(frameNumber%3==0){waterFrame++;}
  if(waterFrame >=8){waterFrame=0;}

}


